#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Persist VLBA metadata extracted from .idifits files
"""

import datetime
import hashlib
import logging
import os
import re
import sys
import time

from pycapo import CapoConfig

from . import NGASUtil as ng
from . import archiveutils, proposalutils
from .schema import create_session
from .schema.model import (
    Author,
    Configuration,
    DataDescription,
    ExecutionBlock,
    File,
    Filegroup,
    Intent,
    Polarization,
    Project,
    Scan,
    ScienceProduct,
    Subscan,
)
from .VLBAFITSUtil import VLBAFits

session = create_session("VLBA")
LOG = logging.getLogger("ingestion")
config = CapoConfig()


def persistVLBAMetadata(fits_file):
    LOG.info(f"Persisting fits file: {fits_file}")
    # Basically want to bail out of the whole process if the file name is already in the Files table
    starttime = time.time()
    vlba_fits = VLBAFits(fits_file)
    LOG.info(f"Fits file completed in {time.time() - starttime}s")
    starttime = time.time()
    projectcode = vlba_fits.getproject()
    segment = vlba_fits.getsegment()

    project = persistproject(projectcode)
    session.add(project)
    # Persist archive location information
    correlation_filegroup = persistfilegroups(project, segment)
    file = persistfile(correlation_filegroup, fits_file)

    # Now the science data
    executionblock = persistexecblock("VLBA", project, segment, correlation_filegroup, vlba_fits, file)
    if executionblock is not None:
        if project.starttime is None or project.starttime > executionblock.starttime:
            project.starttime = executionblock.starttime
        if project.endtime is None or project.endtime < executionblock.endtime:
            project.endtime = executionblock.endtime

    session.add(project)

    LOG.info(f"Database updates for file {executionblock.ngas_fileset_id} completed in {time.time() - starttime}s")
    LOG.info(
        f"Correlation file: {executionblock.ngas_fileset_id} successfully ingested for Project: {executionblock.project_code} Segment: {executionblock.segment}"
    )

    session.commit()


def is_test_project(project_code):
    """
    Is this a test project? Then don't set the proprietary period.
    Test projects follow the pattern: LI*, SI*, CI*, XI*, KI*, UI*, DQ*, PPM*, T*

    In addition there is a handful of test observations with a bit of a one off codes, so I suppose these could be just
    handled as they come (hopefully not too many!). Such three observations from past 2 years are: MK617, PI2010, IPTFEB02

    :param projectcode:
    :return:
    """

    if (
        project_code.startswith("LI")
        or project_code.startswith("SI")
        or project_code.startswith("CI")
        or project_code.startswith("XI")
        or project_code.startswith("KI")
        or project_code.startswith("UI")
        or project_code.startswith("DQ")
        or project_code.startswith("PPM")
        or project_code.startswith("T")
        or project_code == "MK617"
        or project_code == "PI2010"
        or project_code == "IPTFEB02"
    ):
        return True
    else:
        return False


def persistproject(projectcode):

    existingproject = archiveutils.getproject(projectcode)
    if existingproject is None:
        searchproposalcode = re.sub("([A-Z]+)0", "\g<1>", projectcode)
        pstproposal = proposalutils.getproposal("VLBA", searchproposalcode)
        pstauthors = persistauthors(projectcode)
        if pstproposal is None:
            psttitle = ""
            pstabstract = ""
            if is_test_project(projectcode):
                proprietaryduration = 0
            else:
                proprietaryduration = 365
        else:
            if pstproposal.PROPOSAL_TYPE == "Director's Discretionary Time":
                proprietaryduration = round(365 / 2, 0)
            else:
                proprietaryduration = 365
            psttitle = pstproposal.TITLE
            pstabstract = pstproposal.ABSTRACT

        project = Project(
            project_code=projectcode,
            opt_project_id=0,
            legacy_id=projectcode,
            title=psttitle,
            abstract=pstabstract,
            total_observation_time=None,
            last_addition=datetime.datetime.utcnow(),
            starttime=None,
            endtime=None,
            proprietary_duration=proprietaryduration,
        )
        project.authors = pstauthors
        return project
    return existingproject


def persistauthors(project):
    searchproposalcode = re.sub("([A-Z]+)0", "\g<1>", project)
    pstauthors = proposalutils.getauthors("VLBA", searchproposalcode)
    projauthors = []
    if pstauthors is None or len(pstauthors) == 0:
        author = Author(
            project_code=project,
            username="operations",
            firstname="VLBA",
            lastname="Operations",
            pst_person_id=10154,
            is_pi=True,
        )
        session.add(author)
        projauthors.append(author)
    else:
        for pstauthor in pstauthors:
            projauthors.append(
                Author(
                    project_code=project,
                    username=pstauthor.user.personName,
                    firstname=pstauthor.FIRST_NAME,
                    lastname=pstauthor.LAST_NAME,
                    pst_person_id=pstauthor.person_id,
                    is_pi=proposalutils.ispi(pstauthor),
                )
            )

    return projauthors


def persistfilegroups(project, segment):
    filegroup = Filegroup(project_code=project.project_code, datasize=0, type="correlation")
    session.add(filegroup)
    return filegroup


def persistfile(filegroup, fits_file):
    file_path = os.path.basename(fits_file)
    ngas_id = os.path.basename(fits_file)
    existingfile = archiveutils.getvlbafile(ngas_id)
    site_name = config.getstring("archive-ingestion.NGASSite")
    cluster_name = config.getstring("archive-ingestion.NGASCluster")
    if existingfile is None:
        file = File(
            file_path=file_path,
            ngas_id=ngas_id,
            ngas_location=site_name,
            ngas_cluster=cluster_name,
            filegroup1=filegroup,
            filename=os.path.basename(fits_file),
            filesize=os.path.getsize(fits_file),
            format=fits_file.suffix[1:].upper(),
            type=fits_file.suffix[1:].upper(),
            checksum=sha256(fits_file),
            checksum_type="SHA256",
            ingestion_time=datetime.datetime.utcnow(),
        )

        file.filegroup1.datasize = file.filegroup1.datasize + file.filesize
        session.add(file)
        return file
    LOG.critical(f"File {ngas_id} already exists. Exiting.")
    exit(-1)


def persistexecblock(telescope, project, segment, filegroup, vlba_fits, file):
    science_product_locator = ng.generate_science_product_locator(telescope, "correlation")
    current_band_set = set()  # for the file
    execblock = ExecutionBlock(
        ost_exec_block_id=0,
        filegroup=filegroup,
        calibration_level=0,
        telescope=telescope,
        configuration=None,
        scheduling_block_id=0,
        project=project,
        segment=project.project_code + segment,
        ngas_fileset_id=file.filename,
        starttime=0,
        endtime=0,
        calibration_status="Do Not Calibrate",
        scheduling_block_type=None,
        science_product_locator=science_product_locator,
    )

    science_product = ScienceProduct(
        science_product_locator=science_product_locator,
        science_product_type="Correlation",
        metadata_ingestion_date=datetime.datetime.utcnow(),
        metadata_ingestion_version=1,
        filegroup=filegroup,
        external_name=file.filename,
        external_system="VLBA Operations",
        execution_block=execblock,
    )

    execblock.project.science_products.append(science_product)

    session.add(execblock)
    session.add(science_product)

    starttime, endtime, bandset = persistscans(vlba_fits, execblock, filegroup, file)
    execblock.starttime = starttime
    execblock.endtime = endtime

    execblock.band_code = " ".join(sorted(bandset))

    return execblock


def persistscans(vlba_fits, execution_block, filegroup, file):
    minstarttime, maxendtime = 1000000000.0, 0.0
    scans = vlba_fits.getscantable()
    bandset = set()
    for scan_number in range(len(scans)):
        scan = Scan(
            execution_block=execution_block,
            filegroup=filegroup,
            min_bandwidth=0.0,
            max_bandwidth=0.0,
            min_frequency=0.0,
            max_frequency=0.0,
            polarization_code=0,
        )
        minfrequency, maxfrequency, bandwidth, starttime, endtime, polarization = persistsubscans(
            execution_block, scan, scans[scan_number], file
        )

        # MHz --> Hz
        scan.min_frequency = minfrequency * 1000000
        scan.max_frequency = maxfrequency * 1000000

        scan.min_bandwidth = bandwidth
        scan.max_bandwidth = bandwidth
        # scan.polarization_code = polarization.polarization_id
        session.add(scan)
        if starttime < minstarttime:
            minstarttime = starttime
        if endtime > maxendtime:
            maxendtime = endtime

        # get the band for the minfrequency and add to a set
        bandset.add(vlba_fits.lookup_band(minfrequency / 1000))
        bandset.add(vlba_fits.lookup_band(maxfrequency / 1000))

    return minstarttime, maxendtime, bandset


def persistsubscans(execution_block, scan, scandata, file):
    configuration = persistconfiguration(execution_block, 0)
    subscan = Subscan(
        scan=scan,
        obstype="",
        starttime=scandata["START_TIME"],
        endtime=scandata["END_TIME"],
        sourcename=scandata["SOURCE"].split("\x00", 1)[0],  # Fix some source names with terminating nulls
        sourcetype="STAR",
        ra=scandata["RA"],
        dec=scandata["DEC"],
        exposure_time=scandata["TIME_ON_SOURCE"],
        integration_time=scandata["INTEGRATION_TIME"],
        receiver_id=0,
        backend="DIFX",
        configuration=configuration,
        intent="TRACK",
        file=file,
    )

    bandwidth = scandata["BANDWIDTH"]
    frequencies = [float(freq) for freq in scandata["FREQUENCY"].split(",")]
    minfrequency = min(frequencies)
    maxfrequency = max(frequencies)
    polarization = scandata["POLARIZATION"]
    for frequency in frequencies:
        persistdatadescription(subscan, configuration, frequency, bandwidth, polarization)

    session.add(subscan)
    return minfrequency, maxfrequency, bandwidth, subscan.starttime, subscan.endtime, polarization


def persistdatadescription(subscan, configuration, frequency, bandwidth, polarization):
    local_polarization = session.merge(polarization)
    datadescription = DataDescription(
        bandwidth=bandwidth, frequency=frequency, polarization=local_polarization, configuration=configuration
    )
    session.add(datadescription)


def persistconfiguration(execution_block, configuration):
    configuration = Configuration(execution_block=execution_block, configuration=configuration)
    session.add(configuration)
    return configuration


def sha256(fname):
    hash_sha256 = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


if __name__ == "__main__":
    path = sys.argv[1]
    fits_file = sys.argv[2]
    persistVLBAMetadata(os.path.join(path, fits_file))
