#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import json
import requests
from marshmallow import Schema, fields, post_load
from pycapo import CapoConfig
from typing import List

from .interfaces import ProductMetadataFinder, ProductMetadata, FetchPlan, FetchContext, FetchPlanDecorator


class ProductMetadataSchema(Schema):
    """Encapsulates the product metadata"""

    telescope = fields.String()
    project_code = fields.String()
    type = fields.String()
    segment = fields.String(required=False, allow_none=True)
    mous_id = fields.String(required=False, allow_none=True)
    date = fields.Float(required=False, allow_none=True)

    @post_load
    def make_report(self, data, **kwargs):
        return ProductMetadata(**data)


class ArchiveProductMetadataFinder(ProductMetadataFinder):
    def retrieve_metadata(self, product_locator: str) -> ProductMetadata:
        settings = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings")
        response = requests.get(settings.metadataServiceUrlPrefix, params=dict(locator=product_locator))
        return ProductMetadataSchema().load(response.json())


class ProductMetadataFileGenerator(FetchPlanDecorator):
    """
    Fetches the product metadata file for some product locators.
    """

    def __init__(self, other: FetchPlan, locators: List[str]):
        super().__init__(other)
        self.locators = locators

    def fetch(self, context: FetchContext):
        # run the parent
        super().fetch(context)

        # now do our thing
        self.generate_product_metadata_file(context)

    def generate_product_metadata_file(self, context: FetchContext):
        """
        Generates a "products.json" in the delivery root which contains the map of the product metadata.

        :param context: the current fetch context
        """
        result = {}
        finder = ArchiveProductMetadataFinder()

        # for each product, figure out the product location we used for it
        for locator in self.locators:
            # this was the location we used
            product_location = context.locator_fetch_destination(locator).relative_to(context.destination)

            # put the metadata in the result with the location as the key
            result[str(product_location)] = finder.retrieve_metadata(locator).__json__()

        # record the product metadata for all our products
        (context.destination / "products.json").write_text(json.dumps(result))
